/***********************************************************************

  ColorBox2: utils.h
    (c) 2004 by Martin Decky
  
  Distributed under the terms of Lesser General Public Licence, see
  COPYING.

  0.3			10th August 2004	complete rewrite, C++

***********************************************************************/


#ifndef __CB2__UTILS_H__
#define __CB2__UTILS_H__


#include <math.h>
#include <string>
#include <ostream>


/** Generic assert macro.
 *
 * If the condition 'cond' is false,
 * ExceptionAssert is thrown.
 *
 */
#define CB2_ASSERT(cond, warning) \
	if (!(cond)) \
		throw ExceptionAssert(warning ": " #cond " in " __FILE__, __LINE__);
		

#define CB2_RANGECHECK(cond, value) \
	if (!(cond)) \
		throw ExceptionRange< typeof(value)>(#cond ": Value " #value "=", value, " is out of range in " __FILE__, __LINE__);
		
		
#define CB2_MEMORYCHECK(cond) \
	if (!(cond)) \
		throw ExceptionMemory(#cond ": Memory dereference fault in " __FILE__, __LINE__);


/** ColorBox2 namespace
 *
 * All ColorBox2 types and classes
 * are within the CB2 namespace.
 *
 */
namespace CB2 {
	
	
	/** Base class of all ColorBox2 classes
	 *
	 * Common methods and properties.
	 *
	 */
	class Class {
		public:
			size_t Sizeof() const; /**< Size of the class */
	};
	
	
	template< class Tsize> class ExceptionRange; /**< Forward declaration */
	class ExceptionConversion;                   /**< Forward declaration */
	class ExceptionMemory;                       /**< Forward declaration */
	class ExceptionAssert;                       /**< Forward declaration */
	
	
	/** Generic iterator class
	 *
	 * This template implements a generic iterator
	 * of data items from Tcontainer. The iterator
	 * refers to the items by Tref and returns
	 * the values as Tmemb.
	 *
	 * Tcontainer must implement these methods:
	 *
	 *   - Tref RefNext(const Tref &)
	 *       To the given reference generate
	 *       the reference to the next item.
	 *       Should not throw any exceptions
	 *       (the successor of the final or invalid
	 *       reference should be the final reference).
	 *
	 *   - Tmemb Deref(const Tref &)
	 *       Get the value of the referenced
	 *       item. Might throw any useful exception
	 *       on final or invalid reference).
	 *
	 */
	template< class Tcontainer, class Tref, class Tmemb> class Iterator : public Class {
		public:
			
			
			/** Constructor
			 *
			 * Create an iterator refering
			 * to the given item.
			 *
			 * @param cont      Container.
			 * @param reference Reference to an item.
			 *
			 */
			Iterator(const Tcontainer & cont, const Tref & reference)
				: container(cont) {
				ref = reference;
			}
			
			
			/** Advance iterator reference
			 *
			 * Use container's RefNext() to get
			 * the reference to the succeding item.
			 *
			 * @return Reference to this.
			 *
			 */
			const Iterator< Tcontainer, Tref, Tmemb> & operator++() {
				ref = container.RefNext(ref);
				return *this;
			}
			
			
			/** Compare iterators
			 *
			 * Compare iterators by comparing
			 * the references.
			 *
			 * @param it Iterator to compare with.
			 *
			 * @return Compare result.
			 *
			 */
			bool operator==(const Iterator< Tcontainer, Tref, Tmemb> & it) const {
				return (ref == it.ref);
			}
			
			
			/** Compare iterators
			 *
			 * Compare iterators by comparing
			 * the references.
			 *
			 * @param it Iterator to compare with.
			 *
			 * @return Compare result.
			 *
			 */
			bool operator!=(const Iterator< Tcontainer, Tref, Tmemb> & it) const {
				return (ref != it.ref);
			}
			
			
			/** Dereference iterator
			 *
			 * Get the item referenced by the iterator
			 * using container's Deref() method.
			 *
			 * @return Dereferenced value.
			 *
			 */
			Tmemb operator*() const {
				return container.Deref(ref);
			}
			
		private:
			
			const Tcontainer & container; /**< Container reference */
			Tref ref;                     /**< Current item reference */
			
	};
	
	
	/** Multipurpose array with reference counting
	 *
	 * Multipurpose array type for storing values of
	 * the same kind. Grows in fragments, never
	 * shrinks by inself (shrinks only explicitly
	 * on Discart(true) method).
	 *
	 * Note: This is a low-level array class, the elements
	 * are stored and copied bit-by-bit (no copy constructors
	 * or class assignments are called). Use List
	 * for storing class objects.
	 *
	 * NB: As a low-level storage implementation, many
	 * operations have strict limitations on self-reference
	 * (i.e. you shouldn't insert values referenced from
	 * the same array instance). Use List for more robust
	 * (but slower) implementation.
	 *
	 */
	template< class Tmemb> class Array : public Class {
		public:
		
			typedef size_t size_type;                                   /**< Size type */
			typedef Tmemb * ref_type;                                   /**< Item reference type */
			typedef Iterator< Array< Tmemb>, ref_type, Tmemb> iterator; /**< Iterator */
		

			/** Default constructor
			 *
			 * Create empty array.
			 *
			 * @param fragment Default fragment size
			 *                 (in items count).
			 *
			 */
			Array(const size_type fragment = 4096) {
				/* Empty data */
				data = NULL;
				allocated = 0;
				frag = fragment;
				allocate(0);
			}
			
			
			/** Copy constructor
			 *
			 * Copy array items from another array
			 * (copy is bit-by-bit).
			 *
			 * @param prev Array to copy.
			 *
			 */
			Array(const Array< Tmemb> & prev) {
				/* Empty data */
				data = NULL;
				allocated = 0;
				frag = prev.frag;
				privatize(prev);
			}
			
			
			/** Desctructor
			 *
			 * Dispose the memory of the array.
			 *
			 */
			~Array() {
				if (data) /* Dispose data */
					delete[] data;
			}
			
			
			/** Copy assignment
			 *
			 * Copy array items from another array
			 * (copy is bit-by-bit).
			 *
			 * @param prev Array to copy.
			 *
			 */
			const Array< Tmemb> & operator=(const Array< Tmemb> & prev) {
				if (&prev != this) { /* No self-assignment */
					frag = prev.frag;
					privatize(prev);
				}
				return *this;
			}
			
			
			/** Implicit conversion operator
			 *
			 * Returns constant pointer to the array.
			 *
			 * @return Constant pointer.
			 *
			 */
			operator const Tmemb * () const {
				return (const Tmemb * ) data;
			}
			
			
			/** Append item at the end.
			 *
			 * Append one item at the end
			 * of the array.
			 *
			 * @param prev Member to append.
			 *
			 * @return Reference to this.
			 *
			 */
			const Array< Tmemb> & operator<<(const Tmemb & prev) {
				allocate(len + 1);
				memcpy(data + len - 1, &prev, sizeof(Tmemb));
				
				return *this;
			}
			
			
			/** Append item at the beginning.
			 *
			 * Append item at the beginning
			 * of the array.
			 *
			 * Note: The appended item should
			 * not be a reference to another item
			 * from this array (in this case
			 * a wrong item will be appended).
			 *
			 * @param prev Item to append.
			 *
			 * @return Reference to this.
			 *
			 */
			const Array< Tmemb> & operator>>(const Tmemb & prev) {
				allocate(len + 1);
				
				/* Shift original items
				   and put new item at the beginning */
				memmove(data + 1, data, len - 1);
				memcpy(data, &prev, sizeof(Tmemb));
				
				return *this;
			}
			
			
			/** Enlarge array by a number of items.
			 *
			 * Enlarge the array by allocating memory
			 * for 'nmemb' items at the end. The contents
			 * of the items is undefined.
			 *
			 * @param nmemb Number of items to allocate.
			 *
			 * @return Reference to this.
			 *
			 */
			const Array< Tmemb> & operator+=(const size_type nmemb) {
				allocate(len + nmemb);
				return *this;
			}
			
			
			/** Shrink array by a number of items.
			 *
			 * Shrink the array by disposing 'nmemb'
			 * items at the end. If a larger count
			 * of items than the length of the array
			 * is specified, the array made empty.
			 *
			 * @param nmemb Number of items to dispose.
			 *
			 * @return Reference to this.
			 *
			 */
			const Array< Tmemb> & operator-=(const size_type nmemb) {
				if (nmemb > len)
					allocate(0);
				else
					allocate(len - nmemb);
				
				return *this;
			}
			
			
			/** Get item.
			 *
			 * Get the item from the array indexed by 'pos'.
			 * If 'pos' is out of range, ExceptionRange
			 * is thrown.
			 *
			 * @param pos Index of the member.
			 *
			 * @return Copy of the member.
			 *
			 */
			Tmemb operator[](const size_type pos) const {
				CB2_RANGECHECK(pos < len, pos);
				return data[pos];
			}
			
			
			/** Get item reference.
			 *
			 * Get reference to an item from the array
			 * indexed by 'pos'. If 'pos' is out
			 * of range, ExceptionRange is thrown.
			 *
			 * @param pos Index of the member.
			 *
			 * @return Reference to a member.
			 *
			 */
			Tmemb & operator[](const size_type pos) {
				CB2_RANGECHECK(pos < len, pos);
				return data[pos];
			}
			
			
			/** Copy items from C-style array.
			 *
			 * Copy items from a standard C-style
			 * array. Members are copied bit-by-bit.
			 *
			 * Original contents of the array is discarted.
			 *
			 * @param buffer Array with members to copy.
			 * @param nmemb  Number of members to copy.
			 *
			 */
			void Fill(const Tmemb * buffer, const size_type nmemb) {
				allocate(nmemb);
				memmove(data, buffer, sizeof(Tmemb) * nmemb);
			}
			
			
			/** Discart the array.
			 *
			 * Discart the contents of the array.
			 *
			 * @param shrink Whether the memory should
			 *               be disposed.
			 *
			 */
			void Discart(const bool shrink = false) {
				if (shrink) {
					/* Dispose data */
					if (data)
						delete[] data;
					data = NULL;
					allocated = 0;
				}
				allocate(0);
			}
			
			
			/** Return number of items.
			 *
			 * @return Number of items in the array.
			 *
			 */
			size_type Len() const {
				return len;
			}
			
			
			/** Check array for empty.
			 *
			 * @return True if the array is empty.
			 *
			 */
			bool Empty() const {
				return (len == 0);
			}
			
			
			/** Get starting reference
			 *
			 * Get the reference to the first
			 * item in the array as a pointer
			 * to the first item (or NULL on empty
			 * array).
			 *
			 * @return Starting reference.
			 *
			 */
			ref_type RefBegin() const {
				if (Len() > 0)
					return data;
					
				return NULL;
			}
			
			
			/** Get final reference
			 *
			 * Get the final reference as the pointer
			 * to the memory position after the last
			 * item in the array (or NULL on empty
			 * array).
			 *
			 * @return Final reference.
			 *
			 */
			ref_type RefEnd() const {
				size_type len = Len();
				
				if (len > 0)
					return (data + len);
					
				return NULL;
			}
			
			
			/** Get reference to item.
			 *
			 * Get the reference to the item
			 * indexed by 'pos' (or final reference
			 * if the index is out of range).
			 *
			 * @param pos Index of the item.
			 *
			 * @return Reference to the item.
			 *
			 */
			ref_type RefPos(const size_type pos) const {
				if (pos < Len())
					return (data + pos);
					
				return RefEnd();
			}
			
			
			/** Get succeding reference
			 *
			 * Get the reference to an item
			 * succeding the item referenced
			 * by 'prev' (i.e. increase the pointer
			 * by one item). If the pointer is
			 * final or invalid, final reference
			 * is returned).
			 *
			 * @param prev Previous reference.
			 *
			 * @return Next reference.
			 *
			 */
			ref_type RefNext(const ref_type & prev) const {
				if ((prev >= RefEnd()) || (prev < RefBegin())) /* Check pointer range */
					return RefEnd();
				
				return (prev + 1);
			}
			
			
			/** Dereference
			 *
			 * Get the item value referenced by 'ref'.
			 * If the reference is out of range,
			 * ExceptionMemory is thrown.
			 *
			 * @param ref Item reference.
			 *
			 * @return Item value.
			 *
			 */
			Tmemb Deref(const ref_type & ref) const {
				/* Check pointer range */
				CB2_MEMORYCHECK((prev < RefEnd()) && (prev >= RefBegin()));
				return (*ref);
			}
			
			
			/** Get starting iterator
			 *
			 * @return Starting iterator.
			 *
			 */
			iterator Begin() const {
				return iterator(*this, RefBegin());
			}
			
			
			/** Get final iterator
			 *
			 * @return Final iterator.
			 *
			 */
			iterator End() const {
				return iterator(*this, RefEnd());
			}
			
			
			/** Cut out piece of array
			 *
			 * Remove several items from the array.
			 * The items are defined by the position of
			 * the first removed item and count of items
			 * that follows (and should be removed).
			 *
			 * If 'pos' or 'cnt' is out of range,
			 * ExceptionRange is thrown.
			 *
			 * @param pos Index of the first item to remove.
			 * @param cnt Number of items to remove.
			 *
			 */
			void Cut(const size_type pos, const size_type cnt) {
				/* Range check */
				CB2_RANGECHECK(pos < len, pos);
				CB2_RANGECHECK(pos + cnt <= len, pos);
				
				if (cnt > 0) {
					size_type mv = len - pos - cnt;
					
					if (mv > 0) /* Shift items front */
						memmove(data + pos, data + pos + cnt, sizeof(Tmemb) * mv);
					
					/* Discart overlapping items */
					allocate(len - cnt);
				}
			}
			
			
			/** Insert items
			 *
			 * Insert items from another array
			 * into this array. The first item from
			 * 'ins' is inserted at index 'pos'
			 * and totally 'cnt' items are inserted
			 * (previous items on these positions
			 * are shifted back).
			 *
			 * If 'pos' or 'cnt' is out of range,
			 * ExceptionRange is thrown.
			 *
			 * NB: This method doesn't support
			 * self-reference. If 'ins' is this
			 * same array, ExceptionAssert is thrown.
			 *
			 * @param pos Index of the first inserted item.
			 * @param cnt Number of items to insert.
			 * @param ins Array of items to insert.
			 *
			 */
			void Insert(const size_type pos, const size_type cnt, const Array< Tmemb> & ins) {
				CB2_ASSERT(&ins != this, "Cannot insert self-referred items");
			
				/* Range check */
				CB2_RANGECHECK(pos < len, pos);
				CB2_RANGECHECK(cnt < ins.Len(), cnt);
				
				size_t mv = len - pos;

				allocate(len + cnt);
				if (mv > 0) /* Shift items back */
					memmove(data + pos + cnt, data + pos, sizeof(Tmemb) * mv);
				
				/* Insert items */
				memcpy(data + pos, ins.data, sizeof(Tmemb) * cnt);
			}
			
		private:

			Tmemb * data;        /**< Data storage */
			size_type allocated; /**< Allocated memory space */
			size_type len;       /**< Members used */
			size_type frag;      /**< Fragment size */
			
			
			/** Allocate memory for items
			 *
			 * Allocate or reallocate memory
			 * for items.
			 *
			 * The memory is rounded to nearest
			 * fragment. If the new count of items
			 * is lower then the original count,
			 * the memory pool is untouched.
			 *
			 * If the allocation/reallocation
			 * fails, ExceptionMemory is thrown.
			 *
			 * @param nlen Number of items.
			 *
			 */
			void allocate(const size_type nlen) {
				/* Round items to nearest fragment */
				len = nlen;
				size_type nalloc = ((len / frag) + 1) * frag;
					
				if (nalloc > allocated) { /* Enlarge memory pool */
					Tmemb * newdata = new Tmemb[nalloc];
					if (data != NULL) {
						memcpy(newdata, data, sizeof(Tmemb) * allocated);
						delete[] data;
					}
					data = newdata;
					allocated = nalloc;
				}
			}
			
			
			/** Copy data from another array.
			 *
			 * Copy data from another array, unless
			 * this is self-reference.
			 *
			 * @param prev Array to copy from.
			 *
			 */
			void privatize(const Array< Tmemb> & prev) {
				if (&prev != this) {
					allocate(prev.len);
					memcpy(data, prev.data, sizeof(Tmemb) * prev.len);
				}
			}
			
	};
	
	
	template< class Tchar> class PString; /**< Forward declaration */
	
	
	/** Binary concatenation operator
	 *
	 * Concatenate two strings and get
	 * the output string.
	 *
	 * @param left  Left operand.
	 * @param right Right operand.
	 *
	 * @return Concatenated strign.
	 *
	 */
	template< class Tchar> const PString< Tchar> operator+(const PString< Tchar> & left, const PString< Tchar> & right) {
		PString< Tchar> output(left);
		output += right;
		return output;
	}
	
	
	template< class Tchar> std::basic_ostream< Tchar, std::char_traits< Tchar> > & operator<<(std::basic_ostream< Tchar,  std::char_traits< Tchar> > & stream, const PString< Tchar> & str) {
		stream.write(str.data, str.Len());
		return stream;
	}
	
	
	/** String template
	 *
	 * Template for representing character
	 * strings. Supports many features not
	 * even available by STL's string template.
	 *
	 * NB: The character (Tchar) type should
	 * be "primitive" (i.e. char, int, wchar_t, etc.),
	 * because Array< Tchar> is used as underlying
	 * data storage (which stores characters
	 * as bit copies).
	 *
	 */
	template< class Tchar> class PString : public Class {
		public:
		
			typedef typename Array< Tchar>::size_type size_type;          /**< Size type */
			typedef typename Array< Tchar>::ref_type ref_type;            /**< Item reference type */
			typedef Iterator< PString< Tchar>, ref_type, Tchar> iterator; /**< Iterator */
			
			
			/** Default constructor
			 *
			 * Create an empty string.
			 *
			 */
			PString()
				: data(def_frag) {
				/* Add NULL-termination */
				data << 0;
			}
			
			
			/** Copy constructors
			 *
			 * Copy string data from another string
			 * repersentation.
			 *
			 * @param prev String to copy.
			 * @param dt   Raw character data.
			 * @param ln   Length of raw character data.
			 *
			 */
			PString(const PString< Tchar> & prev) /**< From PString */
				: data(prev.data) {
			}
			
			PString(const Tchar * dt, const size_type ln) /**< From raw character data */
				: data(def_frag) {
				data.Fill(dt, ln);
				/* Add NULL-termination */
				data << 0;
			}
			
			PString(const Tchar * prev) /**< From NULL-terminated string */
				: data(def_frag) {
				cpy(prev);
			}
			
			PString(const std::basic_string< Tchar> & prev) /**< From std::basic_string< Tchar> */
				: data(def_frag) {
				cpy(prev);
			}
			
			PString(const Tchar prev) /**< From single Tchar */
				: data(def_frag) {
				cpy(prev);
			}
			
			PString(const signed int prev, const size_type base = 10) /**< From signed int */
				: data(def_frag) {
				cpy_numeric_int(prev, base);
			}
			
			PString(const unsigned int prev, const size_type base = 10)  /**< From unsigned int */
				: data(def_frag) {
				cpy_numeric_int(prev, base);
			}
			
			PString(const signed long int prev, const size_type base = 10)  /**< From signed long int */
				: data(def_frag) {
				cpy_numeric_int(prev, base);
			}

			PString(const unsigned long int prev, const size_type base = 10)  /**< From unsigned long int */
				: data(def_frag) {
				cpy_numeric_int(prev, base);
			}

			PString(const signed long long int prev, const size_type base = 10)  /**< From signed long long int */
				: data(def_frag) {
				cpy_numeric_int(prev, base);
			}

			PString(const unsigned long long int prev, const size_type base = 10)  /**< From unsigned long long int */
				: data(def_frag) {
				cpy_numeric_int(prev, base);
			}
			
			PString(const float prev) /**< From float */
				: data(def_frag) {
				cpy_numeric_float(prev);
			}
			
			PString(const double prev) /**< From double */
				: data(def_frag) {
				cpy_numeric_float(prev);
			}
			
			
			/** Assignment operator
			 *
			 * Copy string data from another string
			 * repersentation.
			 *
			 * @param prev String to copy.
			 *
			 * @return Reference to the result.
			 *
			 */
			const PString< Tchar> & operator=(const PString< Tchar> & prev) { /**< From PString */
				if (&prev != this)
					data = prev.data;
				return *this;
			}
			
			const PString< Tchar> & operator=(const Tchar * prev) { /**< From NULL-terminated string */
				cpy(prev);
				return *this;
			}
			
			const PString< Tchar> & operator=(const std::basic_string< Tchar> prev) { /**< From std::basic_string< Tchar> */
				cpy(prev);
				return *this;
			}
			
			const PString< Tchar> & operator=(const Tchar prev) { /**< From single Tchar */
				cpy(prev);
				return *this;
			}
			
			const PString< Tchar> & operator=(const signed int prev) { /**< From signed int */
				cpy_numeric_int(prev, 10);
				return *this;
			}
			
			const PString< Tchar> & operator=(const unsigned int prev) {  /**< From unsigned int */
				cpy_numeric_int(prev, 10);
				return *this;
			}
			
			const PString< Tchar> & operator=(const signed long int prev) {  /**< From signed long int */
				cpy_numeric_int(prev, 10);
				return *this;
			}
			
			const PString< Tchar> & operator=(const unsigned long int prev) {  /**< From unsigned long int */
				cpy_numeric_int(prev, 10);
				return *this;
			}
			
			const PString< Tchar> & operator=(const signed long long int prev) {  /**< From signed long long int */
				cpy_numeric_int(prev, 10);
				return *this;
			}
			
			const PString< Tchar> & operator=(const unsigned long long int prev) {  /**< From unsigned long long int */
				cpy_numeric_int(prev, 10);
				return *this;
			}
			
			const PString< Tchar> & operator=(const float prev) {  /**< From float */
				cpy_numeric_float(prev);
				return *this;
			}
			
			const PString< Tchar> & operator=(const double prev) {  /**< From double */
				cpy_numeric_float(prev);
				return *this;
			}
			
			
			const PString< Tchar> operator<<(const Tchar & right) {
				data[data.Len() - 1] = right;

				/* Add NULL-termination */
				data << 0;
				
				return *this;
			}
			
			
			/** Concatenation operators
			 *
			 * Unary and binary concatenation
			 * operators
			 *
			 * @param right Right concatenation operand.
			 * @param left  Left concatenation operand.
			 *
			 * @return Reference to the result.
			 *
			 */
			const PString< Tchar> & operator+=(const PString< Tchar> & right) { /**< Unary concatenation (+=) */
				/* Remove NULL-termination from left operand */
				data -= 1;
				
				/* Concatenate characters */
				for (size_type i = 0; i < right.Len(); i++)
					data << right.data[i];
				
				/* Add NULL-termination */
				data << 0;
				
				return *this;
			}
			
			const PString< Tchar> & operator+(const PString< Tchar> & right) { /**< Binary concatenation (+) */
				*this += right;
				return *this;
			}
			
			friend const PString< Tchar> operator+<>(const PString< Tchar> & left, const PString< Tchar> & right); /**< Binary concatenation operator (friend variant) */
			
			
			friend std::basic_ostream< Tchar,  std::char_traits< Tchar> > & operator<<<>(std::basic_ostream< Tchar,  std::char_traits< Tchar> > & stream, const PString< Tchar> & str);
			
			
			/** Implicit conversion operators
			 *
			 * These operators do the implicit
			 * conversion to various data types.
			 *
			 * If the destination type cannot
			 * represent the current string value,
			 * ExceptionConversion is thrown.
			 *
			 * If the range of the destination type
			 * is not sufficient, ExceptionRange
			 * is thrown.
			 *
			 * @return String representation in given type.
			 *
			 */
			operator const Tchar * () const { /**< To NULL-terminated string */
				return (const Tchar * ) data;
			}
			
			operator std::basic_string< Tchar> () const { /**< To std::basic_string< Tchar> */
				std::basic_string< Tchar> str;
			
				for (size_type i = 0; i < Len(); i++)
					str += data[i];
					
				return str;
			}
			
			operator Tchar () const { /**< To single Tchar (first character) */
				if (Len() < 1)
					throw ExceptionConversion("Can't convert empty string to character");
				
				return (Tchar) data[(size_type) 0];
			}
			
			operator signed int () const { /**< To signed int */
				return (signed int) numeric(true, (signed int) 0);
			}
			
			operator unsigned int () const {  /**< To unsigned int */
				return (unsigned int) numeric(false, (unsigned int) 0);
			}
			
			operator signed long int () const {  /**< To signed long int */
				return (signed long int) numeric(true, (signed long int) 0);
			}
			
			operator unsigned long int () const {  /**< To unsigned long int */
				return (unsigned long int) numeric(false, (unsigned long int) 0);
			}
			
			operator signed long long int () const {  /**< To signed long long int */
				return (signed long long int) numeric(true, (signed long long int) 0);
			}
			
			operator unsigned long long int () const {  /**< To unsigned long long int */
				return (unsigned long long int) numeric(false, (unsigned long long int) 0);
			}
			
			operator float () const { /**< To float */
				return (float) numeric(true, (float) 0);
			}

			operator double () const { /**< To double */
				return (double) numeric(true, (double) 0);
			}
			
			
			/** Number of Tchar elements
			 *
			 * @return Length of the string.
			 *
			 */
			size_type Len() const {
				size_type len = data.Len();
				CB2_ASSERT(len > 0, "Incosistent NULL-termination");
				return (len - 1);
			}
			
			
			/** Real size of the string in bytes
			 *
			 * @return Size of bytes representing
			 *         the string.
			 *
			 */
			size_type Size() const {
				return (Len() * sizeof(Tchar));
			}

			
			/** Indexaation operator
			 *
			 * Access to a random character
			 * of the string. If out of range,
			 * ExceptionRange is thrown.
			 *
			 * @param pos Index to access.
			 *
			 * @return Character or a reference
			 *         to character on the given
			 *         position.
			 *
			 */
			Tchar operator[](const size_type pos) const { /**< Read-only */
				CB2_RANGECHECK(pos < Len(), pos);
				return data[pos];
			}
			
			Tchar & operator[](const size_type pos) { /**< Read-write */
				CB2_RANGECHECK(pos < Len(), pos);
				return data[pos];
			}
			
			
			/** Get substring
			 *
			 * Get substring from the current string.
			 * The substring is defined by the index
			 * of the first character and length.
			 *
			 * If 'base' is out of range, ExceptionRange
			 * is thrown. If 'length' is too long
			 * (or contains special value (size_type) -1,
			 * which is default), it is truncated to the
			 * maximal possible length.
			 *
			 * @param base   Index of first character.
			 * @param length Number of characters to copy.
			 *
			 * @return Substring.
			 *
			 */
			PString< Tchar> Substr(const size_type base, const size_type length = (size_type) -1) const {
				/* Range check */
				CB2_RANGECHECK(base < Len(), base);
				
				size_type len;
				
				if ((length == (size_type) -1) || (base + length > Len())) /* Truncate length */
					len = Len() - base;
				else
					len = length;
				
				return PString< Tchar>(data + base, len);
			}
			
			
			/** Check for empty
			 *
			 * @return True if string is empty.
			 *
			 */
			bool Empty() const {
				return (Len() == 0);
			}
			
			
			/** Discart the string.
			 *
			 * Discart the contents of the string.
			 *
			 * @param shrink Whether the memory should
			 *               be disposed.
			 *
			 */
			void Discart(const bool shrink = false) {
				data.Discart(shrink);
				/* Add NULL-termination */
				data << 0;
			}
			
			
			/** Find position of substring
			 *
			 * Get the position of the nearest
			 * appearance of 'needle' beginning
			 * at index 'from'.
			 *
			 * If 'needle' or this string are empty,
			 * false is returned.
			 *
			 * @param needle Substring to search.
			 * @param pos    Position of the appearance.
			 * @param from   Index to search from.
			 *
			 * @return True if substring is found.
			 *
			 */
			bool Pos(const PString< Tchar> & needle, size_type & pos, const size_type from = 0) const {
				if ((Empty()) || (needle.Empty()))
					return false;
			
				size_type ni = 0;
				
				for (size_type i = from; i < Len(); i++) {
					while ((i + ni < Len()) && (data[i + ni] == needle.data[ni])) {
						ni++;
						
						if (ni == needle.Len()) { /* Found match */
							pos = i;
							return true;
						}
					}
					ni = 0;
				}
				
				return false;
			}
			
			/** Substitute substring
			 *
			 * Substitute all appearances of 'needle'
			 * by 'replacement'.
			 *
			 * @param needle      String to search.
			 * @param replacement String to substitute.
			 *
			 * @return Number of substitutes done.
			 *
			 */
			size_type Subst(const PString< Tchar> & needle, const PString< Tchar> & replacement) {
				size_type cnt = 0;
				size_type i = 0;
				size_type pos;
				size_type nlen = needle.Len();
				size_type rlen = replacement.Len();
				
				while (Pos(needle, pos, i)) {
					data.Cut(pos, nlen);                      /* Cut out needle */
					data.Insert(pos, rlen, replacement.data); /* Insert replacement */
					
					i = pos + rlen;
					cnt++;
				}
				
				return cnt;
			}
			
			
			PString< Tchar> HTML() const {
				PString< Tchar> str;
				
				for (size_type i = 0; i < Len(); i++) {
					switch (data[i]) {
						case '&':
							str += "&amp;";
							break;
						case '<':
							str += "&lt;";
							break;
						case '>':
							str += "&gt;";
							break;
						case '"':
							str += "&quot;";
							break;
						default:
							str << data[i];
					}
				}
				
				return str;
			}
			
			
			/** Get starting reference
			 *
			 * Get the reference to the first
			 * character in the string
			 * (as a reference to the array).
			 *
			 * @return Starting reference.
			 *
			 */
			ref_type RefBegin() const {
				return data.RefBegin();
			}
			
			
			/** Get final reference
			 *
			 * Get the final reference. This is
			 * the reference to the last character
			 * in the array (NULL-termination).
			 *
			 * @return Final reference.
			 *
			 */
			ref_type RefEnd() const {
				return data.RefPos(Len());
			}
			
			
			/** Get succeding reference
			 *
			 * Get the reference to a character
			 * succeding the character referenced
			 * by 'prev' (i.e. increase the pointer
			 * by one item). If the pointer is
			 * final or invalid, final reference
			 * is returned).
			 *
			 * @param prev Previous reference.
			 *
			 * @return Next reference.
			 *
			 */
			ref_type RefNext(const ref_type & prev) const {
				if ((prev >= RefEnd()) || (prev < RefBegin())) /* Check ranges */
					return RefEnd();
				
				return data.RefNext(prev);
			}
			
			
			/** Dereference
			 *
			 * Get the character referenced by 'ref'.
			 * If the reference is out of range,
			 * ExceptionMemory is thrown.
			 *
			 * @param ref Item reference.
			 *
			 * @return Item value.
			 *
			 */
			Tchar Deref(const ref_type & ref) const {
				/* Check pointer range */
				CB2_MEMORYCHECK((ref < RefEnd()) && (ref >= RefBegin()));
				return (*ref);
			}
			
			
			/** Get starting iterator
			 *
			 * @return Starting iterator.
			 *
			 */
			iterator Begin() const {
				return iterator(*this, RefBegin());
			}
			
			
			/** Get final iterator
			 *
			 * @return Final iterator.
			 *
			 */
			iterator End() const {
				return iterator(*this, RefEnd());
			}
			
		private:
		
			Array< Tchar> data;                    /**< Data storage */
			static const size_type def_frag = 128; /**< Default array fragment size */

			
			/** Copy string
			 *
			 * Do the conversion from another
			 * data type.
			 *
			 * @param prev Data to convert from.
			 *
			 */
			void cpy(const Tchar * prev) { /**< From NULL-terminated string */
				data.Fill(prev, strlen(prev));
				data << 0;
			}
			
			void cpy(const std::basic_string< Tchar> & prev) { /**< From std::basic_string< Tchar> */
				data.Discart();
				
				for (typename std::basic_string< Tchar>::size_type i = 0; i < prev.length(); i++)
					data << prev[i];
				
				data << 0;
			}
			
			void cpy(const Tchar prev) { /**< From single Tchar */
				data.Fill(&prev, 1);
				data << 0;
			}

			
			/** Copy string from integer numeric data type
			 *
			 * Do the conversion from integer numeric
			 * data type.
			 *
			 * NB: The numeric base is not stored
			 * in the string
			 *
			 * @param prev Numeric value to convert from.
			 * @param base Numeric base.
			 *
			 */
			template < class Tnum> void cpy_numeric_int(const Tnum prev, const size_type base) {
				CB2_ASSERT((base > 1) || (base < 37), "Invalid numeric base");
			
				bool sign;
				Tnum value = prev;
				
				data.Discart();
				
				if (value < 0) {
					sign = true;
					value = -value;
				} else
					sign = false;
				
				while (value > 0) {
					Tnum ord = value % base;
					value = value / base;
					if (ord < 10)
						data >> ((Tnum) '0' + ord);
					else
						data >> ((Tnum) 'A' + ord);
				}
				
				if (sign)
					data >> '-';
				
				data << 0;
			}
			
			
			/** Copy string from floating point numeric data type
			 *
			 * Do the conversion from floating
			 * point numeric data type.
			 *
			 * @param prev Numeric value to convert from.
			 *
			 */
			template < class Tnum> void cpy_numeric_float(const Tnum prev) {
				bool sign;
				Tnum value = prev;
				
				data.Discart();
				
				if (value < 0) {
					sign = true;
					value = -value;
				} else
					sign = false;
					
				Tnum value_floor = floor(prev);
				Tnum value_diff = value_floor - value;
				
				while (value_floor > 0) {
					Tnum oval = value_floor;
					value_floor = floor(value_floor / 10);
					int ord = (int) (oval - value_floor * 10);
					data >> ((int) '0' + ord);
				}
				
				if (sign)
					data >> '-';
				
				bool first = true;
				while (value_diff > 0) {
					value_diff = value_diff * 10;
					int ord = (int) value_diff;
					value_floor = floor(value_diff);
					value_diff = value_diff - value_floor;
					if (first) {
						data << '.';
						first = false;
					}
					data << ((int) '0' + ord);
				}
				
				data << 0;
			}
			
			
			/** Convert to numeric data type
			 *
			 * Convert current string value
			 * to numeric data type.
			 *
			 * @param signed_type Is the numeric type signed?
			 * @param zero        Zero value of the numeric type.
			 *
			 * @return Requested value.
			 *
			 */
			template< class Tnum> Tnum numeric(const bool signed_type, const Tnum zero) const {
				Tnum value = zero;
				int state;
				bool sign = false;
				
				if (signed_type)
					state = 0;
				else
					state = 1;
				
				// FIXME floats
				// FIXME non-decimal representations
				
				for (size_type i = 0; i < Len(); i++) {
					switch (data[i]) {
						case '+':
							switch (state) {
								case 0:
									state = 1;
									break;
								default:
									throw ExceptionConversion();
							}
							break;
						case '-':
							if (!signed_type)
								throw ExceptionConversion();
							switch (state) {
								case 0:
									state = 1;
									sign = true;
									break;
								default:
									throw ExceptionConversion();
							}
							break;
						case '0'...'9':
							switch (state) {
								case 0:
								case 1:
								case 2:
									state = 2;
									// FIXME check for ranges
									value *= 10;
									value += (Tnum) data[i] - (Tnum) '0';
									break;
								default:
									throw ExceptionConversion();
							}
							break;
						case ' ':
						case '\t':
						case 0:
							switch (state) {
								case 0:
									break;
								case 2:
									state = 3;
									break;
								default:
									throw ExceptionConversion();
							}
							break;
						default:
							throw ExceptionConversion();
					}
				}
				
				if (sign)
					value = -value;
				
				if (state < 2)
					throw ExceptionConversion();
					
				return value;
				
			}
			
			
	};
	
	
	template< class Tmemb, int N> class Bucket : public Class {
		public:
		
			typedef size_t size_type;			
			
			/** Default constructor
			 *
			 * Leave components uninitialized.
			 *
			 */
			Bucket() {
			}
			
			
			bool operator==(const Bucket< Tmemb, N> & cpl) const {
				for (size_t i = 0; i < N; i++)
					if (values[i] != cpl.values[i])
						return false;
						
				return true;
			}
			
			
			bool operator!=(const Bucket< Tmemb, N> & cpl) const {
				for (size_t i = 0; i < N; i++)
					if (values[i] != cpl.values[i])
						return true;
				
				return false;
			}
			
			
			Tmemb operator[](const size_type pos) const { /**< Read-only */
				CB2_RANGECHECK(pos < N, pos);
				return values[pos];
			}
			
			Tmemb & operator[](const size_type pos) { /**< Read-write */
				CB2_RANGECHECK(pos < N, pos);
				return values[pos];
			}
		
		private:
			Tmemb values[N];
	};
	
	
	/** Split string to tokens
	 *
	 * This template splits string
	 * into tokens (delimited by another
	 * string), which are then accessible
	 * using iterators.
	 *
	 */
	template< class Tchar> class PStringTokenizer : public Class {
		public:
			
			typedef typename PString< Tchar>::size_type size_type;                            /**< Size type */
			typedef Bucket< size_type, 2> ref_type;                                           /**< Reference type */
			typedef Iterator< PStringTokenizer< Tchar>, ref_type, PString< Tchar> > iterator; /**< Iterator */
			
			
			/** Constructor
			 *
			 * Initialize tokenizer with a string
			 * to split and delimiter.
			 *
			 */
			PStringTokenizer(const PString< Tchar> & data, const PString< Tchar> & delimiter)
				: str(data) {
				delim = delimiter;
			}
			
			
			/** Get starting reference
			 *
			 * Get the reference to the first
			 * token in the string.
			 *
			 * @return Starting reference.
			 *
			 */
			ref_type RefBegin() const {
				ref_type ref;
				ref[pos] = 0;
				
				if (str.Pos(delim, ref[len], 0))
					return ref;
				
				if (str.Len() > 0) {
					ref[len] = str.Len();
					return ref;
				}
				
				return RefEnd();
			}
			
			
			/** Get final reference
			 *
			 * Get the final reference. This is
			 * a reference pointing after
			 * the end of the string.
			 *
			 * @return Final reference.
			 *
			 */
			ref_type RefEnd() const {
				ref_type ref;
				
				ref[pos] = str.Len();
				ref[len] = 0;
			
				return ref;
			}
			
			
			/** Get succeding reference
			 *
			 * Get the reference to a token
			 * succeding the token referenced
			 * by 'prev'.
			 *
			 * @param prev Previous reference.
			 *
			 * @return Next reference.
			 *
			 */
			ref_type RefNext(const ref_type & prev) const {
				ref_type ref;
				
				ref[pos] = prev[pos] + prev[len] + delim.Len();
				if (ref[pos] < str.Len()) {
				
					if (str.Pos(delim, ref[len], ref[pos])) {
						ref[len] -= ref[pos];
						return ref;
					}
					
					ref[len] = str.Len() - ref[pos];
					return ref;
				}
				
				return RefEnd();	
			}
			
			
			/** Dereference
			 *
			 * Get the token (substring) referenced by 'ref'.
			 * For final or invalid references,
			 * Substr() method throws ExceptionRange.
			 *
			 * @param ref Token reference.
			 *
			 * @return Token value.
			 *
			 */
			PString< Tchar> Deref(const ref_type & ref) const {
				return str.Substr(ref[pos], ref[len]);
			}
			
			
			/** Get starting iterator
			 *
			 * @return Starting iterator.
			 *
			 */
			iterator Begin() const {
				return iterator(*this, RefBegin());
			}
			
			
			/** Get final iterator
			 *
			 * @return Final iterator.
			 *
			 */
			iterator End() const {
				return iterator(*this, RefEnd());
			}
		
		private:
			
			static const typename ref_type::size_type pos = 0;
			static const typename ref_type::size_type len = 1;
			const PString< Tchar> & str; /**< Reference to the string */
			PString< Tchar> delim;       /**< Token delimiter */
		
	};
	
	
	typedef PString< char> String;                   /**< PString template specialization */
	typedef PStringTokenizer< char> StringTokenizer; /**< PString template specialization */
	
	
	/** Generic ColorBox2 exception class
	 *
	 * Generic exception, other exceptions are derived
	 * from this class.
	 *
	 */
	class Exception : public Class {
		public:
			Exception();
			Exception(const String what);
			Exception(const String what, const int line);
			
			String What() const;
		private:
			String except;
	};
	
	
	/** Range checking exception
	 *
	 * This exception is thrown mostly in range
	 * check faults, i.e. when the range of a destination
	 * type is unable to carry the assigned value, etc.
	 *
	 */
	template< class Tsize> class ExceptionRange : public Exception {
		public:
			ExceptionRange()
				: Exception() {
			}
			
			ExceptionRange(const String what)
				: Exception(what) {
			}
		
			ExceptionRange(const String what, const int line)
				: Exception(what, line) {
			}
			
			ExceptionRange(const String what, const Tsize value, const String what2, const int line)
				: Exception(what + String(value) + what2, line) {
			}
	};
	
	
	/** Conversion exception
	 *
	 * This exception is thrown when a run-time
	 * conversion from one type to another
	 * is not possible due to input or output
	 * limitations (i.e. when trying to convert
	 * a string not representing an integer to integer).
	 *
	 */
	class ExceptionConversion : public Exception {
		public:
			ExceptionConversion();
			ExceptionConversion(const String what);
			ExceptionConversion(const String what, const int line);
	};
	
	
	class ExceptionMemory : public Exception {
		public:
			ExceptionMemory();
			ExceptionMemory(const String what);
			ExceptionMemory(const String what, const int line);
		private:
			size_t acc;
	};
	
	
	class ExceptionAssert : public Exception {
		public:
			ExceptionAssert();
			ExceptionAssert(const String what);
			ExceptionAssert(const String what, const int line);
	};
	
	
	class ExceptionNotFound : public Exception {
		public:
			ExceptionNotFound();
			ExceptionNotFound(const String what);
			ExceptionNotFound(const String what, const int line);
	};
	
};



#endif
