/***********************************************************************

  ColorBox2: mod_cb2.c
	(c) 2004 by Martin Decky
  
  Distributed under the terms of Lesser General Public Licence, see
  COPYING.

  0.1			13th June 2003		initial
  0.2			20th June 2003		forking and binary running
  0.3			10th August 2004	cleanup, documentation

***********************************************************************/
 

#include <stdio.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>


#include "mod_cb2.h"
#include "../config.h"


module MODULE_VAR_EXPORT cb2_module;


/** Get minimum of size_t
 *
 * Get the smaller from the two values
 * of type size_t.
 *
 * @param a Value A.
 * @param b Value B.
 *
 * @return Minimum from A, B.
 *
 */
inline size_t cb2_size_t_min(const size_t a, const size_t b) {
	if (a < b)
		return a;
		
	return b;
}


/** Get maximum of int
 *
 * Get the larger from the two values
 * of type int.
 *
 * @param a Value A.
 * @param b Value B.
 *
 * @return Maximum from A, B.
 *
 */
inline int cb2_int_max(const int a, const int b) {
	if (a > b)
		return a;
		
	return b;
}


/** Check for substring
 *
 * Check whether a string is a substring.
 *
 * @param haystack The string we search in.
 * @param needle   The string we search for.
 *
 * @return True if 'needle' is a substring of 'haystack'.
 *
 */
bool_t cb2_substring(const char * haystack, const char * needle) {
	if ((haystack == NULL) || (needle == NULL)) /* Not valid strings */
		return false;
		
	if (*needle == 0) /* Empty string is always a substring */
		return true;
		
	while (*haystack != 0) {
		
		/* Non-matching characters */
		while (*haystack != *needle) {
			if (*haystack == 0)
				return false;
			haystack++;
		}
		
		const char * hay = haystack;
		const char * nee = needle;
		
		/* Matching characters */
		while (*hay == *nee) {
			hay++;
			nee++;
			if (*nee == 0)
				return true;
		}
		
		haystack++;
	}

	return false;
}


/** Check for absolute path
 *
 * Check whether a path is absolute by examining
 * the first character. Absolute path is denoted
 * by a slash.
 *
 * @param path Path to examine. NULL string is considered
 *             not absolute.
 *
 * @return True if the path is absolute.
 *
 */
inline bool_t cb2_check_absolute_path(const char * path) {
	if (path == NULL)
		return false;
		
	return (path[0] == '/');
}


/** Check for relative path
 *
 * Check whether a path is relative. A path is considered
 * relative when is not absolute or contains relative
 * directory definition (.. component).
 *
 * @param path Path to examine.
 *
 * @param True if path is relative.
 *
 */
inline bool_t cb2_check_relative_path(const char * path) {
	return (!cb2_substring(path, "/"));
}


/** Complex string creation
 *
 * Allocate memory from pool and create
 * complex string using
 *
 * @param p   Pool to allocate memory from.
 * @param fmt Format string.
 *
 * @return Pointer to the created string.
 *
 */
char * cb2_sprintf(struct pool * p, const char * fmt, ...) {
	va_list ap;

	/* Calculate the length of the output */
	va_start(ap, fmt);
	int len = vsnprintf(NULL, 0, fmt, ap);
	va_end(ap);
	
	/* Check for overflow */
	if (len < INT_MAX)
		len++;
	
	/* Allocate string */	
	char * str = (char * ) ap_palloc(p, sizeof(char) * len);
	
	/* Generate output */
	va_start(ap, fmt);
	vsnprintf(str, len, fmt, ap);
	va_end(ap);
	
	return str;
}


void cb2_htmlprint(request_rec * r, const char * data, const size_t len) {
	size_t i;
	
	for (i = 0; i < len; i++) {
		switch (data[i]) {
			case '<':
				ap_rputs("&lt;", r);
				break;
			case '>':
				ap_rputs("&gt;", r);
				break;
			case '&':
				ap_rputs("&amp;", r);
				break;
			default:
				ap_rwrite(data + i, 1, r);
		}
	}
}


/** Generate internal warning message
 *
 * Generate internal warnging message in form of standard
 * HTML warning page.
 *
 * The order of the warning defines either (0) main warning
 * message (with its own header) or (1) secondary warning message
 * (w/o header).
 *
 * @param request_rec Current server request.
 * @param wstatus     Order of the warning.
 * @param kind        Kind of warning.
 * @param desc        Warning message text.
 *
 */
void cb2_warning(request_rec * r, int * wstatus, const enum warn_t kind, const char * desc) {
	if (*wstatus == 0) { /* First warning message (with header) */
		*wstatus = 1;
		
		/* Generate HTML header, set request timeout */
		r->content_type = "text/html";
		ap_soft_timeout(MODULE_NAME ": sending internal warning", r);
		ap_send_http_header(r);
		
		if (r->header_only) {
			/* Just header request */
		    ap_kill_timeout(r);
		    return;
		}
		
		/* Generate HTML body */
		ap_rputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n", r);
		ap_rputs("<html>\n", r);
		ap_rputs("<head>\n", r);
		ap_rputs(" <title>" MODULE_NAME "</title>\n", r);
		ap_rputs("</head>\n", r);
		ap_rputs("<body>\n", r);
	}

	if ((*wstatus < 2) && (!r->header_only)) { /* Secondary warning messages */
		/* Print message heading according warning kind */
		ap_rputs("<h1>", r);
		switch (kind) {
			case warn_int:
				ap_rputs("Internal error", r);
				break;
			case warn_runtime:
				ap_rputs("Run-time error", r);
				break;
			case warn_perm:
				ap_rputs("Permission violation", r);
				break;
			case warn_exec:
				ap_rputs("Execution error", r);
				break;
			case warn_io:
				ap_rputs("File I/O error", r);
				break;
			default:
				ap_rputs("Assertion failed", r);
		}
		ap_rputs("</h1>\n", r);
		
		if (desc != NULL) /* Print warning message text */
			ap_rprintf(r, "<p>%s</p>\n", desc);
	}
}


int cb2_footer(request_rec * r, int * wstatus) {
	if ((*wstatus < 2) && (!r->header_only)) { /* Page footer */
		*wstatus = 2;
	
		ap_rputs("<hr>\n", r);
		ap_rprintf(r, "<address>%s (running on %s)</address>\n", CB2_DESC, ap_get_server_version());
		ap_rputs("</body>\n", r);
		ap_rputs("</html>\n", r);
		
		/* Page finish, kill timeout */
		ap_kill_timeout(r);
	}
	
	return OK;
}


/** Generate file error message
 *
 * Generate warning message conserning I/O
 * operation fault or unsuccesful file permission
 * check.
 *
 * @param request_rec Current server request.
 * @param wstatus     Order of the warning.
 * @param kind        Kind of warning.
 * @param status      I/O operaration or file permission
 *                    check status.
 *
 */
void cb2_file_error(request_rec * r, int * wstatus, const enum warn_t kind, const char * file, const enum io_status_t status) {
	char * desc;
	
	switch (status) {
		case io_owned:
			desc = "is owned by the server process";
			break;
		case io_wrtbl:
			desc = "is writeable by the server process";
			break;
		case io_wwrtbl:
			desc = "is world writeable";
			break;
		case io_notfound:
			desc = "not found";
			break;
		case io_cantstat:
			desc = "cannot be stated";
			break;
		case io_cantopen:
			desc = "cannot be opened";
			break;
		case io_read:
			desc = "read error";
			break;
		case io_exec:
			desc = "execution failed";
			break;
		default:
			desc = "- assertion failed";
	}
	
	/* Create warning message */
	char * msg = ap_pstrcat(r->pool, ap_escape_html(r->pool, file), " ", desc, NULL);
	cb2_warning(r, wstatus, kind, msg);
}


/** Generate exit error message
 *
 * Generate warning conserning child process
 * malfunction (termination by a signal, etc.).
 *
 * @param request_rec Current server request.
 * @param wstatus     Order of the warning.
 * @param kind        Kind of warning.
 * @param file        Filename conserning the warning.
 * @param desc        Description of the warning.
 * @param exnum       Signal number.
 *
 */
void cb2_exit_error(request_rec * r, int * wstatus, const enum warn_t kind, const char * file, const char * desc, const int exnum) {
	/* Create warning message */
	char * msg = ap_pstrcat(r->pool, ap_escape_html(r->pool, file), " ", desc, " ", cb2_sprintf(r->pool, "%d", exnum), NULL);
	cb2_warning(r, wstatus, kind, msg);
}


/** Generate compiler error message
 *
 * Generate description of the Just-In-Time compiler
 * warning (i.e. the output and error output).
 *
 * The warning is displayed only if wstatus == 0 (because
 * no headers can be sent if wstatus > 0).
 *
 * @param request_rec Current server request.
 * @param wstatus     Order of the warning.
 * @param comp        Just-In-Time compiler name.
 * @param exnum       Exit code.
 * @param stdout      Standard output of JIT.
 * @param stlen       Length of the standard output.
 * @param erout       Error output of JIT.
 * @param erlen       Length of the error output.
 *
 */
void cb2_compiler_error(request_rec * r, int * wstatus, const char * comp, const int exnum, const char * stout, const size_t stlen, const char * erout, const size_t erlen) {
	if (*wstatus == 0) { /* Main error message (with header) */
		*wstatus = 1;
		
		/* Generate HTML header, set request timeout */
		r->content_type = "text/html";
		ap_soft_timeout(MODULE_NAME ": sending compiler warning", r);
		ap_send_http_header(r);
		
		if (r->header_only) {
			/* Just header request */
		    ap_kill_timeout(r);
		    return;
		}
		
		/* Generate HTML body */
		ap_rputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n", r);
		ap_rputs("<html>\n", r);
		ap_rputs("<head>\n", r);
		ap_rputs(" <title>" MODULE_NAME "</title>\n", r);
		ap_rputs("</head>\n", r);
		ap_rputs("<body>\n", r);
	}
	
	if ((*wstatus < 2) && (!r->header_only)) { /* Secondary warning messages */
		/* Print message heading according warning kind */
		ap_rputs("<h1>Just-In-Time Compiler Error</h1>\n", r);
		ap_rprintf(r, "<p>%s %s %d</p>\n", ap_escape_html(r->pool, comp), "exited with error code", exnum);
		
		if (stlen > 0) {
			ap_rputs("<h2>Standard output</h2>\n<pre>", r);
			cb2_htmlprint(r, stout, stlen);
			ap_rputs("</pre>\n", r);
		}
		
		if (erlen > 0) {
			ap_rputs("<h2>Error output</h2>\n<pre>", r);
			cb2_htmlprint(r, erout, erlen);
			ap_rputs("</pre>\n", r);
		}
	}
}


/** Check file permissions
 *
 * Checks whether the given file is not owned by the web server,
 * not writeable by the web server (or world-writeable). 
 *
 * @param fname Filename to check.
 *
 * @return Permission check status.
 *
 */
enum io_status_t cb2_check_permissions(const char * fname) {
	struct stat info;
	
	if (fname) {
		if (stat(fname, &info) == 0) {
			if (info.st_uid == ap_user_id)
				return io_owned;
			if ((info.st_gid == ap_group_id) && (info.st_mode & S_IWGRP))
				return io_wrtbl;
			if (info.st_mode & S_IWOTH)
				return io_wwrtbl;
			if (!S_ISREG(info.st_mode))
				return io_cantopen;
		} else {
			switch (errno) {
				case ENOENT:
				case ENOTDIR:
					return io_notfound;
				default:
					return io_cantstat;
			}
		}
	} else
		return io_notfound;
	
	return io_ok;
}


/** Read string from file descriptor.
 *
 * Reads string from file descriptor into a buffer. The string
 * begins at current position in file descriptor and ends
 * at 0x00 character or at character 'delim'. The length of the
 * string is limited by 'size'.
 *
 * If more characters were read from the file that the the string
 * actually contains (there were characters extra after the string
 * delimiter), the count of these extra characters is returned.
 *
 * If the string in the input is longer than the buffer ('size'),
 * the string in buffer is truncated to 'size' - 1 (the last character
 * is zeroed, thus lost) and all the extra characters from input
 * (until 0x00 or 'delim') are read but forgotten.
 *
 * @param fb    File descriptor to read from.
 * @param buf   Buffer to read to.
 * @param size  Length of the buffer (in chars).
 * @param delim String delimiter.
 *
 * @return Count of extra characters in the buffer or negative
 *         value on error (as by read).
 *
 */
ssize_t cb2_read_string(int fd, char * buf, const size_t size, const char delim) {
	size_t zeropos = size;
	size_t bufpos = 0;
	ssize_t rd;
	
	while ((zeropos == size) && (bufpos < size)) {
		/* Read input */
		rd = read(fd, buf + bufpos, sizeof(char) * (size - bufpos));
		if (rd < 0) {
			if ((errno == EINTR) || (errno == EAGAIN))
				continue;
			return rd; /* I/O error */
		}
		
		size_t i;
		/* Is delimiter present? */
		for (i = bufpos; i < cb2_size_t_min(size, bufpos + rd / sizeof(char)); i++) {
			if ((buf[i] == 0) || (buf[i] == delim)) {
				buf[i] = 0;
				zeropos = i;
				break;
			}
		}
			
		/* Update position */
		bufpos += rd / sizeof(char);
		
		if (rd == 0)
			break; /* EOF */
	}
	
	if (zeropos < size)
		return (bufpos - zeropos - 1);

	/* Range checking and extra characters
	   discaring for the degenerated case
	   (string is longer than the buffer) */
	buf[size - 1] = 0;
	   
	char c;
	do {
		rd = read(fd, &c, 1);
		
		if (rd < 0) {
			if ((errno == EINTR) || (errno == EAGAIN))
				continue;
			return rd; /* I/O error */
		}
		
		if (rd == 0)
			break; /* EOF */
			
	} while ((c != 0) && (c != delim));
	
	return 0;
}


/** Get Module configuration
 *
 * Get module configuration for the scope
 * of the given request.
 *
 * @param r Current server request.
 *
 * @return Configuration of the current request's scope.
 *
 */
static struct excfg_t * cb2_dconfig(request_rec * r) {
	return ((struct excfg_t * ) ap_get_module_config(r->per_dir_config, &cb2_module));
}


/** Set configuration compile flag
 *
 * Set the compile flag for given configuration.
 *
 * @param cmd     Command directive parameters (unused).
 * @param mconfig Configuration to alter.
 *
 * @return Always NULL.
 *
 */
static const char * cb2_cmd_compile(cmd_parms * cmd, void * mconfig) {
	((struct excfg_t * ) mconfig)->compile = 1;
	return NULL;
}


/** Set configuration verbose flag
 *
 * Set the verbose flag for given configuration.
 *
 * @param cmd     Command directive parameters (unused).
 * @param mconfig Configuration to alter.
 *
 * @return Always NULL.
 *
 */
static const char * cb2_cmd_verbose(cmd_parms * cmd, void * mconfig) {
	((struct excfg_t * ) mconfig)->verbose = 1;
	return NULL;
}


/** Set configuration binary directory
 *
 * Set the binary directory for given configuration.
 *
 * @param cmd     Command directive parameters (unused).
 * @param mconfig Configuration to alter.
 * @param path    Path to set.
 *
 * @return Error string or NULL on success.
 *
 */
static const char * cb2_cmd_binpath(cmd_parms * cmd, void * mconfig, const char * path) {
	if (!cb2_check_absolute_path(path))
		return "CB2BinaryPath must contain an absolute path";
	
	((struct excfg_t * ) mconfig)->binpath = path;
	return NULL;
}


/** Set configuration source directory
 *
 * Set the source directory for given configuration.
 *
 * @param cmd     Command directive parameters (unused).
 * @param mconfig Configuration to alter.
 * @param path    Path to set.
 *
 * @return Error string or NULL on success.
 *
 */
static const char * cb2_cmd_srcpath(cmd_parms * cmd, void * mconfig, const char * path) {
	if (!cb2_check_absolute_path(path))
		return "CB2SourcePath must contain an absolute path";
	
	((struct excfg_t * ) mconfig)->srcpath = path;
	return NULL;
}


/** Set configuration Just-In-Time compiler path
 *
 * Set the Just-In-Time compiler directory for given configuration.
 *
 * @param cmd     Command directive parameters (unused).
 * @param mconfig Configuration to alter.
 * @param jit     JIT path.
 *
 * @return Error string or NULL on success.
 *
 */
static const char * cb2_cmd_jit(cmd_parms * cmd, void * mconfig, const char * jit) {
	struct stat info;
	
	if (stat(jit, &info) == 0) {
		if (((info.st_uid == ap_user_id) && (info.st_mode & S_IXUSR)) ||
			((info.st_gid == ap_group_id) && (info.st_mode & S_IXGRP)) ||
			(info.st_mode & S_IXOTH)) {
			
			((struct excfg_t * ) mconfig)->jit = jit;
			return NULL;
		} else
			return "CB2JIT must point to an executable";
	}

	return "CB2JIT not found or cannot be stated";
}


/** Run Just-In-Time compiler
 *
 *
 */
int cb2_jit(request_rec * r, int * wstatus, const char * bin_file) {
	struct excfg_t * dcfg = cb2_dconfig(r); /* Local scope configuration */
	int retstatus = OK;                     /* Function return status */
	int out_pipe[2];                        /* JIT stdout pipe */
	int err_pipe[2];                        /* JIT stderr pipe */
	
	/* Allocate buffers for stdout and stderr
	   (the buffers have fixed size, but the size
	   is long enought for any reasonable
	   compiler output and error messages) */
	char * outbuf = (char * ) ap_palloc(r->pool, BUFFER_BYTES);
	char * errbuf = (char * ) ap_palloc(r->pool, BUFFER_BYTES);
	
	/* Number of charactets in buffers */
	size_t outlen = 0;
	size_t errlen = 0;
	
	/* Create communication pipes */
	if (pipe(out_pipe) != 0) {
		cb2_warning(r, wstatus, warn_runtime, "<tt>pipe(out_pipe)</tt> failed in cb2_jit");
		return !OK;
	}
	if (pipe(err_pipe) != 0) {
		/* Cleanup */
		close(out_pipe[0]);
		close(out_pipe[1]);
		cb2_warning(r, wstatus, warn_runtime, "<tt>pipe(err_pipe)</tt> failed in cb2_jit");
		return !OK;
	}
	
	/* Fork a subprocess */
	pid_t pid = fork();
	if (pid < 0) {
		/* Cleanup */
		close(out_pipe[0]);
		close(out_pipe[1]);
		close(err_pipe[0]);
		close(err_pipe[1]);
		cb2_warning(r, wstatus, warn_runtime, "<tt>fork()</tt> failed in cb2_jit");
		return !OK;
	}
	
	if (pid == 0) { /* Child process */
		/* Close default stdout and stderr */
		if (close(1) != 0)
			exit(ret_io);
		if (close(2) != 0)
			exit(ret_io);
		
		/* Open one side of the pipes
		   as stdout and stderr */
		if (dup2(out_pipe[1], 1) == -1)
			exit(ret_io);
		if (dup2(err_pipe[1], 2) == -1)
			exit(ret_io);
			
		/* Cleanup original pipes' file
		   descriptors */
		if (close(out_pipe[0]) != 0)
			exit(ret_io);
		if (close(out_pipe[1]) != 0)
			exit(ret_io);
		if (close(err_pipe[0]) != 0)
			exit(ret_io);
		if (close(err_pipe[1]) != 0)
			exit(ret_io);
			
	    /* Execute the compiler */
		if (execl(dcfg->jit, dcfg->jit, dcfg->srcpath, dcfg->binpath, bin_file, NULL) == -1)
			exit(ret_exec);
		
		exit(ret_ok);
	} else { /* Parent process */
		/* Close the other end
		   of the pipes */
		int state = 0;
		if (close(out_pipe[1]) == 0)
			state |= 1;
		if (close(err_pipe[1]) == 0)
			state |= 2;
			
		if ((state & 3) == 3) {
			
			int out_eof = false;
			int err_eof = false;
			/* Read stdout and stderr */
			while ((!out_eof) || (!err_eof)) {
				fd_set rfds;
				struct timeval tv;
				
				FD_ZERO(&rfds);
				FD_SET(out_pipe[0], &rfds);
				FD_SET(err_pipe[0], &rfds);
				tv.tv_sec = 1;
				tv.tv_usec = 0;
				
				int ret = select(cb2_int_max(out_pipe[0], err_pipe[0]) + 1, &rfds, NULL, NULL, &tv);
				if (ret > 0) {
					if (FD_ISSET(out_pipe[0], &rfds)) { /* Data on stdout */
						ssize_t rd = read(out_pipe[0], outbuf + outlen * sizeof(char), (BUFFER_SIZE - outlen) * sizeof(char));
						if (rd < 0) {
							if ((errno == EINTR) || (errno == EAGAIN))
								continue;
							cb2_warning(r, wstatus, warn_exec, "pipe broken");
							retstatus = !OK;
							break;
						} else if (rd > 0) {
							outlen += rd / sizeof(char);
							// FIXME buffer full
						} else
							out_eof = true;
					}
					
					if (FD_ISSET(err_pipe[0], &rfds)) { /* Data on stderr */
						ssize_t rd = read(err_pipe[0], errbuf + errlen * sizeof(char), (BUFFER_SIZE - errlen) * sizeof(char));
						if (rd < 0) {
							if ((errno == EINTR) || (errno == EAGAIN))
								continue;
							cb2_warning(r, wstatus, warn_exec, "pipe broken");
							retstatus = !OK;
							break;
						} else if (rd > 0) {
							errlen += rd / sizeof(char);
							// FIXME buffer full
						} else
							err_eof = true;
					}
				}
			
			}
				
			/* Wait for child to finish */
			int status;
			pid_t pd = waitpid(pid, &status, 0);
			
			/* Check for child return value */
			if (pd < 0) {
				cb2_warning(r, wstatus, warn_runtime, "<tt>waitpid()</tt> failed in cb2_jit");
				retstatus = !OK;
			} else {
				if (WIFEXITED(status)) {
					switch (WEXITSTATUS(status)) {
						case ret_ok:
							break;
						case ret_io:
							retstatus = !OK;
							cb2_warning(r, wstatus, warn_int, "Child I/O error in cb2_jit");
							break;
						case ret_exec:
							retstatus = !OK;
							cb2_warning(r, wstatus, warn_int, "Child execution error in cb2_jit");
							break;
						default:
							retstatus = !OK;
							if (dcfg->verbose)
								cb2_compiler_error(r, wstatus, dcfg->jit, WEXITSTATUS(status), outbuf, outlen, errbuf, errlen);
					}
				} else if (WIFSIGNALED(status)) {
					retstatus = !OK;
					if (dcfg->verbose)
						cb2_exit_error(r, wstatus, warn_exec, dcfg->jit, "terminated by signal", WTERMSIG(status));
					else
						cb2_exit_error(r, wstatus, warn_exec, "Just-In-Time compiler", "terminated by signal", WTERMSIG(status));
				} else if (WIFSTOPPED(status)) {
					retstatus = !OK;
					if (dcfg->verbose)
						cb2_exit_error(r, wstatus, warn_exec, dcfg->jit, "stopped by signal", WTERMSIG(status));
					else
						cb2_exit_error(r, wstatus, warn_exec, "Just-In-Time compiler", "stopped by signal", WTERMSIG(status));
				} else {
					retstatus = !OK;
					cb2_warning(r, wstatus, warn_int, "<tt>waitpid()</tt> unknown return status in cb2_jit");
				}
			}
		}
		
		/* Cleanup */
		close(out_pipe[0]);
		if ((state & 1) != 1)
			close(out_pipe[1]);
		close(err_pipe[0]);
		if ((state & 2) != 2)
			close(err_pipe[1]);
		
	}
	
	return retstatus;
}


/** Handle request for ColorBox2 file
 *
 * Handle the request for ColorBox2 file.
 *
 * @param r Request.
 *
 * @return OK on succes.
 *
 */
static int cb2_handler(request_rec * r) {
	struct excfg_t * dcfg = cb2_dconfig(r); /* Local scope configuration */
	int wstatus = 0;                        /* Status of the request */
	
	/* Check security permissions
	   of the meta-file */
	enum io_status_t perm = cb2_check_permissions(r->filename);
	
	if (perm != io_ok) {
		cb2_file_error(r, &wstatus, warn_perm, r->uri, perm);
		return cb2_footer(r, &wstatus);
	}
	
	/* Open meta-file */
	int meta = ap_popenf(r->pool, r->filename, O_RDONLY, 0);
	if (meta == -1) {
		cb2_file_error(r, &wstatus, warn_io, r->uri, io_cantopen);
		return cb2_footer(r, &wstatus);
	}
	
	/* Allocate single request buffer
	   for various purposes */
	char * buf = (char * ) ap_palloc(r->pool, BUFFER_BYTES);
	
	/* Read meta-description from meta-file,
	   create absolute path */
	if (cb2_read_string(meta, buf, BUFFER_SIZE, '\n') < 0) {
		cb2_file_error(r, &wstatus, warn_io, r->uri, io_read);
		return cb2_footer(r, &wstatus);
	}
	
	char * bin_path = ap_pstrcat(r->pool, dcfg->binpath, "/", buf, NULL);
	
	/* Check whether the binary file
	   is up-to-date */
	if (dcfg->compile) {
		if (cb2_jit(r, &wstatus, buf) != OK)
			return cb2_footer(r, &wstatus);
	}
		
	/* Check security permissions
	   of the binary file */
	perm = cb2_check_permissions(bin_path);
	if (perm != io_ok) {
		cb2_file_error(r, &wstatus, warn_perm, bin_path, perm);
		return cb2_footer(r, &wstatus);
	}
	
	int req_pipe[2]; /* Request pipe */
	int res_pipe[2]; /* Response pipe */
	
	/* Create communication pipes */
	if (pipe(req_pipe) != 0) {
		cb2_warning(r, &wstatus, warn_runtime, "<tt>pipe(req_pipe)</tt> failed in handler");
		return cb2_footer(r, &wstatus);
	}

	if (pipe(res_pipe) != 0) {
		/* Cleanup */
		close(req_pipe[0]);
		close(req_pipe[1]);
		cb2_warning(r, &wstatus, warn_runtime, "<tt>pipe(res_pipe)</tt> failed in handler");
		return cb2_footer(r, &wstatus);
	}
	
	/* Fork a subprocess */	    
	pid_t pid = fork();
	if (pid < 0) {
		/* Cleanup */
		close(res_pipe[0]);
		close(res_pipe[1]);
		close(req_pipe[0]);
		close(req_pipe[1]);
		cb2_warning(r, &wstatus, warn_runtime, "<tt>fork()</tt> failed in handler");
		return cb2_footer(r, &wstatus);
	}
	
	if (pid == 0) { /* Child process */
		/* Close default stdin and stdout */
		if (close(0) != 0)
			exit(ret_io);
		if (close(1) != 0)
			exit(ret_io);
			
		/* Open one side of the pipes
		   as stdin and stdout */
		if (dup2(req_pipe[0], 0) == -1)
			exit(ret_io);
		if (dup2(res_pipe[1], 1) == -1)
			exit(ret_io);
			
		/* Cleanup original pipes' file
		   descriptors */
		if (close(req_pipe[0]) != 0)
			exit(ret_io);
		if (close(req_pipe[1]) != 0)
			exit(ret_io);
		if (close(res_pipe[0]) != 0)
			exit(ret_io);
		if (close(res_pipe[1]) != 0)
			exit(ret_io);	
		
	    /* Execute the binary */
		if (execl(bin_path, bin_path, r->uri, r->args, NULL) == -1)
			exit(ret_exec);
		
		exit(ret_ok);
	} else { /* Parent process */
		/* Close the other end
		   of the pipes */
		int state = 0;
		if (close(res_pipe[1]) == 0)
			state |= 1;
		if (close(req_pipe[0]) == 0)
			state |= 2;
			
		if ((state & 3) == 3) {
			/* Write input to the request pipe */
			// FIXME write input (ssize_t) written = write(req_pipe[1], buf, size);
			
			/* Read content type from the response pipe
			   (content type is delimited from the body by \n) */
			ssize_t rd = cb2_read_string(res_pipe[0], buf, BUFFER_SIZE, '\n');
			if (rd < 0)
				cb2_warning(r, &wstatus, warn_exec, "unable to get content-type");
			else {
				wstatus = 1;
				r->content_type = ap_pstrdup(r->pool, buf);
				
				ap_soft_timeout("mod_cb2: sending application data", r);
				ap_send_http_header(r);
				
				if ((!r->header_only) && (rd > 0)) { /* Some response data already in the buffer */
					size_t len = strlen(buf);
					ap_rwrite(buf + len + 1, sizeof(char) * rd, r);
				}
				
				/* Read response from the pipe */
				do  {
					rd = read(res_pipe[0], buf, BUFFER_BYTES);
					
					if (rd < 0) {
						if ((errno == EINTR) || (errno == EAGAIN))
							continue;
						cb2_warning(r, &wstatus, warn_exec, "pipe broken");
						break;
					}
				
					if ((!r->header_only) && (rd > 0))
						ap_rwrite(buf, rd, r);
				} while (rd > 0);
			}
				
			// FIXME what if timeout occurs? should the child be killed?
			   
			/* Wait for child to finish */
			int status;
			pid_t pd = waitpid(pid, &status, 0);
			
			/* Check for child return value */
			if (pd < 0)
				cb2_warning(r, &wstatus, warn_runtime, "<tt>waitpid()</tt> failed in handler");
			else {
				if (WIFEXITED(status)) {
					switch (WEXITSTATUS(status)) {
						case ret_ok:
							break;
						case ret_io:
							cb2_warning(r, &wstatus, warn_exec, "child I/O error");
							break;
						case ret_exec:
							if (dcfg->verbose)
								cb2_file_error(r, &wstatus, warn_exec, bin_path, io_exec);
							else
								cb2_file_error(r, &wstatus, warn_exec, "Binary file", io_exec);
							break;
						default:
							if (dcfg->verbose)
								cb2_exit_error(r, &wstatus, warn_exec, bin_path, "ended with error code", WEXITSTATUS(status));
					}
				} else if (WIFSIGNALED(status)) {
					if (dcfg->verbose)
						cb2_exit_error(r, &wstatus, warn_exec, bin_path, "terminated by signal", WTERMSIG(status));
					else
						cb2_exit_error(r, &wstatus, warn_exec, "Binary file", "terminated by signal", WTERMSIG(status));
				} else if (WIFSTOPPED(status)) {
					if (dcfg->verbose)
						cb2_exit_error(r, &wstatus, warn_exec, bin_path, "stopped by signal", WTERMSIG(status));
					else
						cb2_exit_error(r, &wstatus, warn_exec, "Binary file", "stopped by signal", WTERMSIG(status));
				} else
					cb2_warning(r, &wstatus, warn_int, "<tt>waitpid()</tt> unknown return status in handler");
			}
		}
		
		/* Cleanup */
		close(res_pipe[0]);
		if ((state & 1) != 1)
			close(res_pipe[1]);
		if ((state & 2) != 2)
			close(req_pipe[0]);
		close(req_pipe[1]);
	}
	
	return OK;
}


/** Create directory configuration
 *
 * Create directory specific configuration.
 *
 * @param p       Resource pool.
 * @param dirspec Directory to which the configuration is specific.
 *
 * @return Generated configuration.
 *
 */
static void * cb2_create_dir_config(pool * p, char * dirspec) {
	struct excfg_t * cfg = (struct excfg_t * ) ap_pcalloc(p, sizeof(struct excfg_t));

	cfg->compile = 0;
	cfg->binpath = "/var/cb2/bin";
	cfg->srcpath = "/var/cb2/src";
	cfg->jit = INSTALL_PREFIX "/bin/cb2jit";
	cfg->inherited = 0;
	cfg->cmode = cm_directory;
	cfg->loc = ap_pstrcat(p, "DIR(", ((dirspec != NULL) ? dirspec : ""), ")", NULL);
	
	return ((void * ) cfg);
}


/** Merge directory configurations
 *
 * @param p           Resource pool.
 * @param parent_conf Parent configuration.
 * @param newloc_conf New configuration.
 *
 * @return Merged configuration.
 *
 */
static void * cb2_merge_dir_config(pool * p, void * parent_conf, void * newloc_conf) {
	struct excfg_t * merged_config = (struct excfg_t * ) ap_pcalloc(p, sizeof(struct excfg_t));

	merged_config->compile = ((struct excfg_t * ) newloc_conf)->compile;
	merged_config->binpath = ((struct excfg_t * ) newloc_conf)->binpath;
	merged_config->srcpath = ((struct excfg_t * ) newloc_conf)->srcpath;
	merged_config->jit = ((struct excfg_t * ) newloc_conf)->jit;
	merged_config->loc = ap_pstrdup(p, ((struct excfg_t * ) newloc_conf)->loc);
	merged_config->inherited = (((struct excfg_t * ) parent_conf)->inherited | ((struct excfg_t * ) parent_conf)->compile);
	merged_config->cmode = (((struct excfg_t * ) parent_conf)->cmode == ((struct excfg_t * ) newloc_conf)->cmode) ? ((struct excfg_t * ) parent_conf)->cmode : cm_combo;

	return ((void * ) merged_config);
}


/** Create server-wide configuration
 *
 * Create server-wide configuration.
 *
 * @param p Resource pool.
 * @param s Server description (hostname, etc.).
 *
 * @return Generated configuration.
 *
 */
static void * cb2_create_server_config(pool * p, server_rec * s) {
	struct excfg_t * cfg = (struct excfg_t * ) ap_pcalloc(p, sizeof(struct excfg_t));
	char * sname = s->server_hostname;

	cfg->compile = 0;
	cfg->binpath = "/var/cb2/bin";
	cfg->srcpath = "/var/cb2/src";
	cfg->jit = INSTALL_PREFIX "/bin/cb2jit";
	cfg->inherited = 0;
	cfg->cmode = cm_server;
	cfg->loc = ap_pstrcat(p, "SVR(", ((sname != NULL) ? sname : ""), ")", NULL);

	return ((void * ) cfg);
}


/** Merge server-wide configurations
 *
 * @param p            Resource pool.
 * @param server1_conf Parent configuration.
 * @param server2_conf New configuration.
 *
 * @return Merged configuration.
 *
 */
static void * cb2_merge_server_config(pool * p, void * server1_conf, void * server2_conf) {
	struct excfg_t * merged_config = (struct excfg_t * ) ap_pcalloc(p, sizeof(struct excfg_t));
	
	merged_config->compile = ((struct excfg_t * ) server2_conf)->compile;
	merged_config->binpath = ((struct excfg_t * ) server2_conf)->binpath;
	merged_config->srcpath = ((struct excfg_t * ) server2_conf)->srcpath;
	merged_config->jit = ((struct excfg_t * ) server2_conf)->jit;
	merged_config->loc = ap_pstrdup(p, ((struct excfg_t * ) server2_conf)->loc);
	merged_config->inherited = (((struct excfg_t * ) server1_conf)->inherited | ((struct excfg_t * ) server1_conf)->compile);
	merged_config->cmode = (((struct excfg_t * ) server1_conf)->cmode == ((struct excfg_t * ) server2_conf)->cmode) ? ((struct excfg_t * ) server1_conf)->cmode : cm_combo;

	return((void * ) merged_config);
}


/** Configuration directives description
 *
 * Structure describing mod_cb2 configuration
 * directives and callbacks.
 *
 */
static const command_rec cb2_cmds[] = {
	{
		"CB2Compile",                                       /**< Directive name */
		cb2_cmd_compile,                                    /**< Config action routine */
		NULL,                                               /**< Argument to include in call */
		OR_OPTIONS,                                         /**< Where available */
		NO_ARGS,                                            /**< Arguments */
		"Enable just-in-time compilation and recompilation" /**< Directive description */
	},
	{
		"CB2Verbose",
		cb2_cmd_verbose,
		NULL,
		OR_OPTIONS,
		NO_ARGS,
		"Verbose error messages"
	},
	{
		"CB2BinaryPath",
		cb2_cmd_binpath,
		NULL,
		OR_OPTIONS,
		TAKE1,
		"ColorBox2 binaries path"
	},
	{
		"CB2SourcePath",
		cb2_cmd_srcpath,
		NULL,
		OR_OPTIONS,
		TAKE1,
		"ColorBox2 sources path"
	},
	{
		"CB2JIT",
		cb2_cmd_jit,
		NULL,
		OR_OPTIONS,
		TAKE1,
		"ColorBox2 Just-In-Time compiler"
	},
	{NULL}
};


/** Module handlers
 *
 * MIME handlers that mod_cb2 processes
 * and proper callbacks.
 *
 */
static const handler_rec cb2_handlers[] = {
	{
		"application/x-cb2",
		cb2_handler
	},
	{NULL}
};


/** Module description structure
 *
 * General mod_cb2 description and callbacks.
 *
 */
module MODULE_VAR_EXPORT cb2_module = {
	STANDARD_MODULE_STUFF,
	NULL,                     /**< Module initialization */
	cb2_create_dir_config,    /**< Default directory configuration */
	cb2_merge_dir_config,     /**< Merge current directory configuration with default */
	cb2_create_server_config, /**< Default server-wide configuration */
	cb2_merge_server_config,  /**< Merge current server-wide configuration with default */
	cb2_cmds,                 /**< Configuration directives */
	cb2_handlers,             /**< Module served handlers */
	NULL,                     /**< module_translate_handler */
	NULL,                     /**< module_check_user_id */
	NULL,                     /**< module_auth_checker */
	NULL,                     /**< module_access_checker */
	NULL,                     /**< module_type_checker */
	NULL,                     /**< module_fixer_upper */
	NULL,                     /**< module_loggerr */
	NULL,                     /**< module_header_parser */
	NULL,                     /**< module_child_initr */
	NULL,                     /**< module_child_exit */
	NULL                      /**< module_post_read_request */
};
